unit AbortThreads;

interface

uses
  Windows, Classes;

type
  TAbortThread = class(TThread)
  protected
    procedure Execute; override;
  public
    TargetEID: LongWord;
    hEndEvent: THandle;

    class function BeginAbortWait(NewEID: LongWord): TAbortThread; virtual;
    procedure EndAbortWaitAndFree; virtual;

    constructor Create(CreateSuspended: Boolean); virtual;
    destructor Destroy; override;
    procedure LetEnd; virtual;
  end;

implementation

uses
  AIOUSB;

function GetDeviceHandle(DeviceIndex: LongWord; pData: PHandle): LongWord; cdecl; external 'AIOUSB.dll';

const
  FILE_DEVICE_UNKNOWN = $0022;
  METHOD_BUFFERED = 0;
  METHOD_NEITHER  = 3;
  FILE_ANY_ACCESS = 0;

  IOCTL_ADAPT_INDEX = $0000;
  IOCTL_CY_BASE = LongWord(FILE_DEVICE_UNKNOWN shl 16) or (FILE_ANY_ACCESS shl 14);
  IOCTL_ADAPT_SEND_EP0_CONTROL_TRANSFER = IOCTL_CY_BASE or ((IOCTL_ADAPT_INDEX+8) shl 2) or METHOD_BUFFERED;
  IOCTL_ADAPT_SELECT_INTERFACE          = IOCTL_CY_BASE or ((IOCTL_ADAPT_INDEX+3) shl 2) or METHOD_BUFFERED;
  IOCTL_ADAPT_ABORT_PIPE                = IOCTL_CY_BASE or ((IOCTL_ADAPT_INDEX+17) shl 2) or METHOD_BUFFERED;
  IOCTL_ADAPT_SEND_NON_EP0_DIRECT       = IOCTL_CY_BASE or ((IOCTL_ADAPT_INDEX+18) shl 2) or METHOD_NEITHER;
  IOCTL_ADAPT_GET_DRIVER_VERSION = IOCTL_CY_BASE or ((IOCTL_ADAPT_INDEX) shl 2) or METHOD_BUFFERED;

{ TAbortThread }

class function TAbortThread.BeginAbortWait(NewEID: LongWord): TAbortThread;
begin
  Result := Self.Create(True);
  Result.TargetEID := NewEID;
  Result.Resume;
end;

constructor TAbortThread.Create(CreateSuspended: Boolean);
begin
  inherited;

  hEndEvent := CreateEvent(nil, True, False, nil);
end;

destructor TAbortThread.Destroy;
begin
  CloseHandle(hEndEvent);

  inherited;
end;

procedure TAbortThread.EndAbortWaitAndFree;
begin
  FreeOnTerminate := True;
  LetEnd;
end;

procedure TAbortThread.Execute;
var
  DI, BytesReturned: LongWord;
  hDevice: THandle;
  TargetPipe: Byte;
  BCControlData: LongWord;

begin
  if WaitForSingleObject(hEndEvent, 256) = WAIT_OBJECT_0 then Exit;

  DI := GetDeviceByEEPROMByte(TargetEID);
  if DI <> diNone then begin
    if GetDeviceHandle(DI, @hDevice) = ERROR_SUCCESS then begin
      TargetPipe := $86;
      BytesReturned := 0;
      DeviceIoControl(hDevice, IOCTL_ADAPT_ABORT_PIPE, @TargetPipe, SizeOf(TargetPipe), nil, 0, BytesReturned, nil);

      if WaitForSingleObject(hEndEvent, 256) = WAIT_OBJECT_0 then Exit;
    end;

    BCControlData := $01000007;
    GenericVendorWrite(DI, $BC, 0, 0, SizeOf(BCControlData), @BCControlData);

    if WaitForSingleObject(hEndEvent, 256) = WAIT_OBJECT_0 then Exit;

    BCControlData := $01000007;
    GenericVendorWrite(DI, $BC, 0, 0, SizeOf(BCControlData), @BCControlData);
  end;

  WaitForSingleObject(hEndEvent, INFINITE);
end;

procedure TAbortThread.LetEnd;
begin
  SetEvent(hEndEvent);
end;

end.
